from .types import RedisType


class RedisDao(object):
    """ base class for Redis DAO """

    def __init__(self, redis_client, key_params=None):
        self._key_params = key_params or {}
        self._redis_client = redis_client
        self._set_instance_attributes()

    def __repr__(self):
        return '{}(key_params={})'.format(
            self.__class__.__name__, self._key_params)

    def pipeline(self):
        """ create a Redis pipeline """
        pipeline = self._redis_client.pipeline()
        members = {key: val.clone(redis_client=pipeline)
                   for key, val in self._members.items()}
        return RedisPipeline(pipeline, members)

    def _set_instance_attributes(self):
        """ set instance attributes using class attributes """
        members = {}
        definitions = self._gather_definitions()
        for key, def_ in definitions.items():
            member = def_.clone(
                redis_client=self._redis_client,
                key_params=self._key_params)
            setattr(self, key, member)
            members[key] = member
        self._members = members

    def _gather_definitions(self):
        """ gather RedisType definitions """
        definitions = {}
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if isinstance(attr, RedisType):
                definitions[attr_name] = attr
        return definitions


class RedisPipeline(object):
    """ Redis pipeline """

    def __init__(self, pipeline, members):
        self._pipeline = pipeline
        self._set_members(members)

    def __repr__(self):
        return '{}'.format(self.__class__.__name__)

    def _set_members(self, members):
        self._members = members
        for key, val in members.items():
            setattr(self, key, val)

    def execute(self):
        self._pipeline.execute()



import functools
import logging

import redis


logger = logging.getLogger(__name__)


def ignore_redis_error(default=None):
    """ decorator for ignoring redis.RedisError

    :param default: default return value when RedisError is raised. It also can
                    be callable, which will be used as a factory for creating a
                    default return value.
    """
    def _ignore_redis_error(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except redis.RedisError:
                logger.warn("operation failed on redis", exc_info=True)
                ret = default() if callable(default) else default
                return ret
        return wrapper
    return _ignore_redis_error



""" Define wrappers for each Redis type """
import copy
import functools
import json
from abc import ABCMeta


def _is_key_matured(key, key_params):
    """ key_params provides full information for key or not """
    try:
        key.format(**key_params)
    except KeyError:
        return False
    else:
        return True


class RedisType(object):
    """ Base class of Redis types """

    __metaclass__ = ABCMeta
    SUPPORTED_OPERATIONS = [
        'debug_object', 'delete', 'dump', 'exists', 'expire', 'expireat',
        'move', 'persist', 'pexpire', 'pexpireat', 'pttl', 'rename',
        'renamenx', 'restore', 'ttl', 'type']
    REDIS_TYPE = None

    def __init__(self, key, redis_client=None, key_params=None):
        self._key = key
        self._redis_client = redis_client
        self._key_params = key_params or {}
        self._matured = _is_key_matured(self._key, self._key_params)
        self._configure_attributes()

    def clone(self, **kwargs):
        """ create a copy of myself """
        key = kwargs.get('key', self._key)
        redis_client = kwargs.get('redis_client', self._redis_client)
        key_params = kwargs.get('key_params', self._key_params)
        return self.__class__(key, redis_client, key_params)

    @property
    def key(self):
        return self._key.format(**self._key_params)

    def __repr__(self):
        return '{}(key="{}", key_params={})'.format(
            self.__class__.__name__, self._key, self._key_params)

    def __call__(self, **kwargs):
        if not kwargs:
            return self
        key_params = copy.copy(self._key_params)
        key_params.update(kwargs)
        return self.clone(key_params=key_params)

    def __getattr__(self, name):
        if name in self.SUPPORTED_OPERATIONS:
            # raise useful error if not matured
            raise AttributeError(
                "Not enough keys are provided for redis operation")
        raise AttributeError

    def _configure_attributes(self):
        if self._matured and self._redis_client is not None:
            for operation in self.SUPPORTED_OPERATIONS:
                func = getattr(self._redis_client, operation)
                method = functools.partial(func, self.key)
                # docstring や help を元の関数からコピー
                functools.update_wrapper(method, func)
                setattr(self, operation, method)

    def is_matured(self):
        """ whether key_params support all parameters in key or not """
        return self._matured

    def is_type_consistent(self):
        """ whether type is consistent or not """
        if not self._matured:
            raise RuntimeError(
                "Only matured data is supported for consistency check")
        return self.type() == self.REDIS_TYPE


class RedisString(RedisType):
    """ Redis String type """
    SUPPORTED_OPERATIONS = RedisType.SUPPORTED_OPERATIONS + [
        'append', 'bitcount', 'bitpos', 'decr', 'get', 'getbit', 'getrange',
        'getset', 'incr', 'incrby', 'incrbyfloat', 'lock', 'psetex', 'set',
        'setbit', 'setex', 'setnx', 'setrange', 'strlen', 'substr']
    REDIS_TYPE = b'string'

    def get_json(self):
        """ get and decode as json """
        value = self.get()
        if value is not None:
            return json.loads(value.decode('utf-8'))
        return None

    def set_json(self, value):
        """ encode to json and set to redis """
        value = json.dumps(value, ensure_ascii=False)
        return self.set(value)

    def setex_json(self, seconds, value):
        """ encode to json and setex to redis """
        value = json.dumps(value, ensure_ascii=False)
        return self.setex(seconds, value)


class RedisList(RedisType):
    """ Redis List type """
    SUPPORTED_OPERATIONS = RedisType.SUPPORTED_OPERATIONS + [
        'blpop', 'brpop', 'brpoplpush', 'lindex', 'linsert', 'llen', 'lpop',
        'lpush', 'lpushx', 'lrange', 'lrem', 'lset', 'ltrim', 'rpop',
        'rpoplpush', 'rpush', 'rpushx', 'sort']
    REDIS_TYPE = b'list'

    def lrange_mget(self, start, stop):
        """ mget items by keys obtained by lrange """
        keys = self.lrange(start, stop)
        if not keys:
            return [], []
        return keys, self._redis_client.mget(list(keys))

    def lrange_mget_json(self, start, stop):
        """ mget items by keys obtained by lrange and decode as json """
        keys, values = self.lrange_mget(start, stop)
        items = []
        for x in values:
            if x is not None:
                items.append(json.loads(x.decode('utf-8')))
            else:
                items.append(None)
        return keys, items


class RedisSet(RedisType):
    """ Redis Set type """
    SUPPORTED_OPERATIONS = RedisType.SUPPORTED_OPERATIONS + [
        'sadd', 'scard', 'sdiff', 'sdiffstore', 'sinter', 'sinterstore',
        'sismember', 'smembers', 'smove', 'sort', 'spop', 'srandmember',
        'srem', 'sscan', 'sscan_iter', 'sunion', 'sunionstore']
    REDIS_TYPE = b'set'

    def smembers_mget(self):
        """ mget items by keys obtained by smembers """
        keys = self.smembers()
        if not keys:
            return [], []
        return keys, self._redis_client.mget(list(keys))

    def smembers_mget_json(self):
        """ mget items by keys obtained by smembers and decode as json """
        keys, values = self.smembers_mget()
        items = []
        for x in values:
            if x is not None:
                items.append(json.loads(x.decode('utf-8')))
            else:
                items.append(None)
        return keys, items


class RedisHash(RedisType):
    """ Redis Hash type """
    SUPPORTED_OPERATIONS = RedisType.SUPPORTED_OPERATIONS + [
        'hdel', 'hexists', 'hget', 'hgetall', 'hincrby', 'hincrbyfloat',
        'hkeys', 'hlen', 'hmget', 'hmset', 'hscan', 'hscan_iter', 'hset',
        'hsetnx', 'hvals']
    REDIS_TYPE = b'hash'


class RedisSortedSet(RedisType):
    """ Redis Sorted Set type """
    SUPPORTED_OPERATIONS = RedisType.SUPPORTED_OPERATIONS + [
        'sort', 'zadd', 'zcard', 'zcount', 'zincrby', 'zinterstore',
        'zlexcount', 'zrange', 'zrangebylex', 'zrangebyscore', 'zrank', 'zrem',
        'zremrangebylex', 'zremrangebyrank', 'zremrangebyscore', 'zrevrange',
        'zrevrangebylex', 'zrevrangebyscore', 'zrevrank', 'zscan',
        'zscan_iter', 'zscore', 'zunionstore']
    REDIS_TYPE = b'zset'


# Extended types

# class RedisGeohash(RedisSortedSet):
#     """ Redis Geohash using Scored Set type """
#     SUPPORTED_OPERATIONS = RedisSortedSet.SUPPORTED_OPERATIONS + [
#         'geoadd', 'geodist', 'geohash', 'geopos', 'georadius',
#         'georadiusbymember']


class RedisHyperLogLog(RedisString):
    SUPPORTED_OPERATIONS = RedisString.SUPPORTED_OPERATIONS + [
        'pfadd', 'pfcount', 'pfmerge']



from setuptools import setup, find_packages

version = '0.3.0'

try:
    with open('README.rst') as f:
        readme = f.read()
except IOError:
    readme = ''

setup(
    name="gxredis",
    version=version,
    description="Simple redis-py wrapper library",
    long_description=readme,
    url="https://github.com/groove-x/gxredis",
    author="Junya Hayashi",
    author_email="junya.hayashi@groove-x.com",
    maintainer="Junya Hayashi",
    maintainer_email="junya-hayashi@groove-x.com",
    keywords=['Redis', 'key-value store'],
    packages=find_packages(exclude=["test"]),
    install_requires=["redis"],
    tests_require=["pytest"],
    license="MIT",
    test_suite="test",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Database',
    ]
)



import pytest as pytest
import redis

from gxredis import (
    RedisDao, RedisString, RedisList, RedisHash, RedisSet, RedisSortedSet,
)

HOSTNAME = "localhost"
PORT = 6379
DB = 15


class SampleDao(RedisDao):
    item = RedisString(key="sample")
    item_list = RedisList(key="sample_list")
    item_set = RedisSet(key="sample_set")
    item_hash = RedisHash(key="sample_hash")
    item_zset = RedisSortedSet(key="sample_zset")


class ImmatureDao(RedisDao):
    item = RedisString(key="sample:{item_id}")
    item_list = RedisList(key="sample_list")


def test_dao():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = SampleDao(client)

    dao.item.set("abc")
    assert dao.item.get() == b"abc"
    assert dao.item.is_type_consistent()

    dao.item_list.lpush("abc")
    assert dao.item_list.is_type_consistent()
    dao.item_list.lpop()

    dao.item_set.sadd("abc")
    assert dao.item_set.is_type_consistent()
    dao.item_set.srem("abc")

    dao.item_hash.hset('key', 1)
    assert dao.item_hash.is_type_consistent()
    dao.item_hash.hdel('key')

    dao.item_zset.zadd({'a': 1})
    assert dao.item_zset.is_type_consistent()
    dao.item_zset.zrem('a')


def test_params():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client)

    assert dao.item(item_id=1).key == "sample:1"
    dao.item(item_id=1).set("xyz")
    assert dao.item(item_id=1).get() == b"xyz"
    assert dao.item(item_id=1).is_type_consistent()
    dao.item(item_id=1).delete("xyz")


def test_params_provided_by_dao():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client, key_params={"item_id": 1})

    dao.item.set("xyz")
    assert dao.item.get() == b"xyz"
    assert dao.item.is_type_consistent()
    dao.item.delete()


def test_pipeline():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = SampleDao(client)

    dao.item_list.delete()

    pipe = dao.pipeline()
    pipe.item_list.lpush("a")
    pipe.item_list.lpush("b")
    assert dao.item_list.llen() == 0
    pipe.execute()
    assert dao.item_list.llen() == 2


def test_not_matured():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client)
    with pytest.raises(AttributeError):
        dao.item.get()


def test_attribute_error():
    """ Test for RedisType.__getattr__ """
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client)
    with pytest.raises(AttributeError):
        _ = dao.item.foo


def test_json():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client, key_params={"item_id": 10})

    value = {"hello": "world"}
    dao.item.set_json(value)
    assert dao.item.get() == b'{"hello": "world"}'
    assert dao.item.get_json() == value


def test_lrange_mget():
    client = redis.StrictRedis(HOSTNAME, PORT, DB)
    dao = ImmatureDao(client)

    dao.item_list.delete()
    accr1 = dao.item(item_id=1)
    accr2 = dao.item(item_id=2)
    accr1.set("a")
    accr2.set("b")
    dao.item_list.lpush(accr1.key)
    dao.item_list.lpush(accr2.key)
    keys, items = dao.item_list.lrange_mget(0, 100)

    assert len(items) == 2
    assert keys, [b"sample:2" == b"sample:1"]
    assert items, [b"b" == b"a"]
